Option Strict Off
Option Explicit On

Imports System.Collections.Generic
Imports System.Media
Imports System.Threading

Module Mario

    Declare Function GetKeyState Lib "user32" Alias "GetAsyncKeyState" (ByVal vKey As Integer) As Short
    Declare Function mciSendString Lib "winmm.dll" Alias "mciSendStringA" (ByVal lpstrCommand As String, ByVal lpstrReturnString As String, ByVal uReturnLength As Integer, ByVal hwndCallback As Integer) As Integer

    Const BW As Short = 32
    Const BH As Short = 32
    Const PixelGo As Short = 8

    ' The duration of a tick in seconds
    Const TickDuration As Double = 1 / 30.0

    Delegate Function IntegerExtractor(state As GameState) As Integer
    Delegate Function DoubleExtractor(state As GameState) As Double

    Class GameState
        Public ReadOnly objects() As Objekter
        Public ReadOnly player As Player
        Public ReadOnly offsetX As Integer
        Public ReadOnly offsetY As Integer

        Public Sub New(objects() As Objekter, player As Player, offsetX As Integer, offsetY As Integer)
            Me.objects = objects
            Me.player = player
            Me.offsetX = offsetX
            Me.offsetY = offsetY
        End Sub
    End Class

    Structure Player
        Dim Tag As Short
        Dim Vei As Boolean
        Dim XX As Double
        Dim YY As Double
        Dim Ok As Boolean
        Dim Level As Short
        Dim Liv As Short
        Dim Hjerter As Short
        Dim MaxH As Short
        Dim X As Double
        Dim Y As Double
        Dim Uddelig As Single
        Dim Hopp As Single
    End Structure

    Structure Objekter
        Dim Enabled As Boolean
        Dim X As Double
        Dim Y As Double
        Dim Vei As Short
        Dim Tag As Short
        Dim A As Single
    End Structure

    Private Enum NextGameState
        Running = 0
        Restart = 1
    End Enum

    Public Const vbDstInvert As Integer = &H550009
    Public Const vbMergeCopy As Integer = &HC000CA
    Public Const vbMergePaint As Integer = &HBB0226
    Public Const vbNotSrcCopy As Integer = &H330008
    Public Const vbNotSrcErase As Integer = &H1100A6
    Public Const vbPatCopy As Integer = &HF00021
    Public Const vbPatInvert As Integer = &H5A0049
    Public Const vbPatPaint As Integer = &HFB0A09
    Public Const vbSrcAnd As Integer = &H8800C6
    Public Const vbSrcCopy As Integer = &HCC0020
    Public Const vbSrcErase As Integer = &H440328
    Public Const vbSrcInvert As Integer = &H660046
    Public Const vbSrcPaint As Integer = &HEE0086

    Dim YYYY, Penger, XXXX, XY As Short
    Dim XXX, YYY As Integer
    Dim Objekter_Renamed(999) As Objekter
    Dim Player_Renamed As Player
    Dim State As Double
    Dim Bokser(,) As String
    Dim H As Short
    Dim E(1) As Short
    Dim ABFG As Short

    ' Level has just been loaded
    Dim LevelJustLoaded As Boolean

    Public Pic1 As System.Drawing.Image
    Public GameOver, Igjen As Boolean
    Public C As Integer
    Public Sound As Boolean

    Public ExitGame As Boolean

    Public Const SColor As Short = 24

    ' Cache of sounds (to prevent slowdowns)
    Public soundCache As New Dictionary(Of String, Boolean)

    Sub Game()

        On Error GoTo A
        Dim XX, F, B, A, G, YY As Short

        Sound = True
        Igjen = False
        GameOver = True

        Dim screenBuffer As New Bitmap(CInt(ScreenWidth), CInt(ScreenHeight), Imaging.PixelFormat.Format32bppPArgb)

BegynnIgjen:
        Start()

        Dim bufferWidth As Integer = 0
        Dim bufferHeight As Integer = 0
        Dim buffer As BufferedGraphics = Nothing

        Using target As Graphics = Form1.CreateGraphics()
            Dim prevState As GameState = Nothing
            Dim prevTick As Double = Double.NaN

            Dim currentTick As Double = 0
            Dim simulatedTicks As Double = -1

            Dim timeDelta As New Stopwatch()
            timeDelta.Start()

            GameOver = False
            If Igjen = True Then Igjen = False : GoTo Spillet

            Player_Renamed.Level = 1
            Player_Renamed.Liv = 5
            Player_Renamed.MaxH = 3
            Player_Renamed.Hjerter = 1
            Player_Renamed.X = 0
            Player_Renamed.Y = 0
            Player_Renamed.XX = 0
            Player_Renamed.YY = 0
            LoadLevel(Nothing, Player_Renamed.Level)

            Debug.Print("Started")

            If Sound = True Then
                mciSendString("open " & Chr(34) & GetResourcePath("bomber.mid") & Chr(34), CStr(0), 0, 0)
                mciSendString("play " & Chr(34) & GetResourcePath("bomber.mid") & Chr(34), CStr(0), 0, 0)
            End If

            Do While Not ExitGame
Spillet:
                ' Detect screen resize - allocate new
                If buffer Is Nothing OrElse bufferWidth <> CInt(ScaledWidth) OrElse bufferHeight <> CInt(ScaledHeight) Then
                    If buffer IsNot Nothing Then
                        buffer.Dispose()
                    End If
                    buffer = context.Allocate(target, New Rectangle(0, 0, ScaledWidth, ScaledHeight))
                    bufferWidth = CInt(ScaledWidth)
                    bufferHeight = CInt(ScaledHeight)

                    Debug.Print("Reallocating buffer context")
                End If

                If LevelJustLoaded Then
                    ' TIme delta has been corrupted by the level loading time
                    timeDelta.Reset()
                    LevelJustLoaded = False

                    ' Also reset prev state
                    prevTick = Double.NaN
                End If

                Dim deltaTime As Double = timeDelta.Elapsed.TotalSeconds
                Dim deltaTicks As Double = deltaTime / TickDuration

                ' Limit deltaTicks when debugger is attached
                If Debugger.IsAttached AndAlso deltaTicks > 1 Then
                    deltaTicks = 1
                End If

                currentTick += deltaTicks

                ' The state of the previous tick - this is used for linear interpolation
                If Double.IsNaN(prevTick) OrElse Int(prevTick) <> Int(currentTick) Then
                    ' Save current state
                    prevState = New GameState(Objekter_Renamed.Clone(), Player_Renamed, XXX, YYY)
                    prevTick = currentTick
                End If

                ' Reset and go to the next round
                timeDelta.Reset()
                timeDelta.Start()

                State = State + deltaTicks
                If State >= 9 Then State = 0
                G = G + 1

                If Not Second(TimeOfDay) = F Then H = G : G = 0 : F = Second(TimeOfDay) : C = C + 1
                If C = 36 And Sound = True Then
                    C = 0
                    If Sound Then
                        mciSendString("play " & Chr(34) & GetResourcePath("bomber.mid") & Chr(34) & " from 0", CStr(0), 0, 0)
                    End If
                End If

                ' Draw to background image
                Using screenGraphics = Graphics.FromImage(screenBuffer)
                    screenGraphics.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor

                    DrawAll(screenGraphics, prevState, GetGameState(), currentTick - prevTick)

                    ' Only simulate one tick at a time
                    If Int(currentTick) > Int(simulatedTicks) Then
                        simulatedTicks += 1

                        Dim nextState As NextGameState = DoTicks(screenGraphics, 1)

                        Select Case nextState
                            Case NextGameState.Restart
                                GoTo BegynnIgjen
                            Case NextGameState.Running
                                ' NOP
                        End Select
                    End If

                    ' Draw to background buffer
                    buffer.Graphics.Transform = ScaledTransform
                    buffer.Graphics.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
                    buffer.Graphics.DrawImageUnscaled(screenBuffer, 0, 0)
                    buffer.Render(target)
                End Using

                System.Windows.Forms.Application.DoEvents()
                If Not Speed = 0 Then Thread.Sleep(Speed)
            Loop
            Exit Sub
        End Using

        If buffer IsNot Nothing Then
            buffer.Dispose()
        End If
A:
        Form1.Hide()
        System.Windows.Forms.Application.DoEvents()
        MsgBox("Feil: " & Err.Number & ", " & ErrorToString(), MsgBoxStyle.Critical, "FEIL!")
        End

    End Sub

    Private Function DoTicks(buffer As Graphics, simulatedTicks As Double) As NextGameState
        Dim simulatedDelta As Integer = 1
        simulatedTicks += 1

        Debug.Print("Simulated tick: " & simulatedTicks & ", X: " & Player_Renamed.X & ", Y: " & Player_Renamed.Y)

        ' Move player character
        If Player_Renamed.X + XXX > ScreenWidth / 2 And Int(Player_Renamed.X / BW) < (E(0) - 9) Then XXX = XXX - ((Player_Renamed.X + XXX) - (ScreenWidth / 2))
        If Player_Renamed.X + XXX < ScreenWidth / 2 And Int(Player_Renamed.X / BW) > 9 Then XXX = XXX - ((Player_Renamed.X + XXX) - (ScreenWidth / 2))

        If Player_Renamed.Y + YYY > ScreenHeight / 2 And Int(Player_Renamed.Y / BH) < E(1) - 9 Then YYY = YYY - ((Player_Renamed.Y + YYY) - (ScreenHeight / 2))
        If Player_Renamed.Y + YYY < ScreenHeight / 2 And Int(Player_Renamed.Y / BH) > 7 Then YYY = YYY - ((Player_Renamed.Y + YYY) - (ScreenHeight / 2))

        ' Wrap around
        If Player_Renamed.Uddelig > 0 Then Player_Renamed.Uddelig = Player_Renamed.Uddelig - simulatedDelta
        If Penger >= 100 Then Penger = Penger - 100 : Player_Renamed.Liv = Player_Renamed.Liv + 1

        If GetKeyState(System.Windows.Forms.Keys.Escape) < 0 Then
            If Sound Then mciSendString("close " & Chr(34) & GetResourcePath("bomber.mid") & Chr(34), CStr(0), 0, 0)
            GameOver = False
            Return NextGameState.Restart
        End If

        If GetKeyState(System.Windows.Forms.Keys.Left) < 0 And Player_Renamed.X > 0 And Colisjon(buffer, Player_Renamed.X - PixelGo * simulatedDelta, Player_Renamed.Y, True, "Left") = 0 Then
            Player_Renamed.X = Player_Renamed.X - PixelGo * simulatedDelta
            Player_Renamed.Vei = True
            Debug.Print("New X: " & Player_Renamed.X)
        End If
        If GetKeyState(System.Windows.Forms.Keys.Right) < 0 And Colisjon(buffer, Player_Renamed.X + PixelGo * simulatedDelta, Player_Renamed.Y, True, "Right") = 0 Then
            Player_Renamed.X = Player_Renamed.X + PixelGo * simulatedDelta
            Player_Renamed.Vei = False
            Debug.Print("New X: " & Player_Renamed.X)
        End If
        If GetKeyState(System.Windows.Forms.Keys.Up) < 0 And Player_Renamed.Ok = True Then
            Player_Renamed.Hopp = 10
            Player_Renamed.Ok = False
        End If
        If GetKeyState(System.Windows.Forms.Keys.Up) >= 0 And Not Player_Renamed.Hopp = 0 Then
            Player_Renamed.Hopp = 0
        End If
        If Player_Renamed.Liv = 0 Then GameOver = True : Igjen = False : Return NextGameState.Restart
        If GameOver = True Then Igjen = False : Return NextGameState.Restart

        If Player_Renamed.Hopp <= 0 Then
            Dim D As Short = Colisjon(buffer, Player_Renamed.X, Player_Renamed.Y + PixelGo * simulatedDelta, True, "Down")

            If D = 0 Then
                Player_Renamed.Y = Player_Renamed.Y + PixelGo * simulatedDelta
                Player_Renamed.Ok = False
                Debug.Print("New Y: " & Player_Renamed.Y)
            Else
                Player_Renamed.Ok = True
            End If

            ' Align player to the nearest box
            If D = 1 Then
                Player_Renamed.Y = Math.Round(Player_Renamed.Y / CDbl(BH)) * BH
            End If
            If D = 2 Then
                Return NextGameState.Running
            End If
        End If

        If Player_Renamed.Hopp > 0 Then
            If Colisjon(buffer, Player_Renamed.X, Player_Renamed.Y, True, "Up") = 0 Then
                Player_Renamed.Y = Player_Renamed.Y - PixelGo * simulatedDelta
                Player_Renamed.Hopp = Player_Renamed.Hopp - simulatedDelta
            Else
                Player_Renamed.Hopp = 0
                Dunk()
            End If
        End If


        If Int(Player_Renamed.Y / BH) > E(1) Then
            Player_Renamed.Liv = Player_Renamed.Liv - 1
            LoadLevel(buffer, Player_Renamed.Level)

            Debug.Print("Player killed")
        End If

        For A As Integer = 0 To 999
            If Objekter_Renamed(A).Enabled = True And Objekter_Renamed(A).X + 636 > 0 - XXX And Objekter_Renamed(A).X - 36 < (0 - XXX) + 1280 Then
                If DoObjekts(buffer, A, simulatedDelta) = 2 Then Return NextGameState.Running
            End If
        Next

        Return NextGameState.Running
    End Function

    Function GetGameState()
        Return New GameState(Objekter_Renamed, Player_Renamed, XXX, YYY)
    End Function

    Function DoObjekts(ByVal g As Graphics, ByRef A As Short, elapsedTicks As Double) As Short

        Dim B, XX, YY As Integer
        Dim C As Double

        If Objekter_Renamed(A).Tag = 0 Or Objekter_Renamed(A).Tag = 2 Or Objekter_Renamed(A).Tag = 3 Or Objekter_Renamed(A).Tag = 5 Or Objekter_Renamed(A).Tag = 6 Or Objekter_Renamed(A).Tag = 8 Or Objekter_Renamed(A).Tag = 9 Then
            XX = Objekter_Renamed(A).X / BW
            YY = Objekter_Renamed(A).Y / BH

            If Objekter_Renamed(A).Tag = 5 Then C = (elapsedTicks * PixelGo) / 1.2 Else : C = elapsedTicks * PixelGo
            If Objekter_Renamed(A).Vei = 0 Then

                If Colisjon(g, Objekter_Renamed(A).X - (C * 2), Objekter_Renamed(A).Y, False, "Left") = 0 Then Objekter_Renamed(A).X = Objekter_Renamed(A).X - C Else Objekter_Renamed(A).Vei = 1
            End If

            If Objekter_Renamed(A).Vei = 1 Then
                If Colisjon(g, Objekter_Renamed(A).X + (C * 2), Objekter_Renamed(A).Y, False, "Right") = 0 Then Objekter_Renamed(A).X = Objekter_Renamed(A).X + C Else Objekter_Renamed(A).Vei = 0
            End If

            If Colisjon(g, Objekter_Renamed(A).X, Objekter_Renamed(A).Y + PixelGo * elapsedTicks, False, "Down") = 0 Then
                Objekter_Renamed(A).Y = Objekter_Renamed(A).Y + PixelGo * elapsedTicks
            End If
            If XX <= 0 Or YY < 0 Or XX > E(0) + 1 Or YY + 1 > E(1) Then Objekter_Renamed(A).Enabled = False : Exit Function

            If Bokser(XX - 1, YY + 1) = "A" Or Bokser(XX - 1, YY + 1) = "U" Then If Not Objekter_Renamed(A).Tag = 5 And Not Objekter_Renamed(A).Tag = 6 And Not Objekter_Renamed(A).Tag = 8 And Not Objekter_Renamed(A).Tag = 9 Then Objekter_Renamed(A).Vei = 1
            If Bokser(XX + 1, YY + 1) = "A" Or Bokser(XX + 1, YY + 1) = "U" Then If Not Objekter_Renamed(A).Tag = 5 And Not Objekter_Renamed(A).Tag = 6 And Not Objekter_Renamed(A).Tag = 8 And Not Objekter_Renamed(A).Tag = 9 Then Objekter_Renamed(A).Vei = 0

            If COLLIDE(Player_Renamed.X, Player_Renamed.Y, Objekter_Renamed(A).X, Objekter_Renamed(A).Y, BW, "A") = 1 Then
                If Player_Renamed.Y < Objekter_Renamed(A).Y - 7 Then
                    If Objekter_Renamed(A).Tag = 0 Then Objekter_Renamed(A).Tag = 1 : Player_Renamed.Hopp = 15 : Player_Renamed.Uddelig = 10 : Player_Renamed.Ok = False
                    If Objekter_Renamed(A).Tag = 2 Then Objekter_Renamed(A).Enabled = False
                    If Objekter_Renamed(A).Tag = 3 Then Objekter_Renamed(A).Tag = 4 : Player_Renamed.Hopp = 15 : Player_Renamed.Uddelig = 10 : Player_Renamed.Ok = False
                    If Objekter_Renamed(A).Tag = 5 Then PlussPLiv() : Objekter_Renamed(A).Enabled = False
                    If Objekter_Renamed(A).Tag = 6 Then TaFraLiv(g) : Objekter_Renamed(A).Enabled = False
                    If Objekter_Renamed(A).Tag = 8 Then Player_Renamed.Liv = Player_Renamed.Liv + 1 : Objekter_Renamed(A).Enabled = False
                    If Objekter_Renamed(A).Tag = 9 Then PlussPMax() : Objekter_Renamed(A).Enabled = False
                Else
                    If Objekter_Renamed(A).Tag = 5 Then PlussPLiv() : Objekter_Renamed(A).Enabled = False : GoTo ABC
                    If Objekter_Renamed(A).Tag = 6 Then TaFraLiv(g) : Objekter_Renamed(A).Enabled = False : GoTo ABC
                    If Objekter_Renamed(A).Tag = 8 Then Player_Renamed.Liv = Player_Renamed.Liv + 1 : Objekter_Renamed(A).Enabled = False : GoTo ABC
                    If Objekter_Renamed(A).Tag = 9 Then PlussPMax() : Objekter_Renamed(A).Enabled = False : GoTo ABC
                    If Player_Renamed.Uddelig <= 0 Then Player_Renamed.Hjerter = Player_Renamed.Hjerter - 1
                    Player_Renamed.Uddelig = 50

                    If Player_Renamed.Hjerter = 0 Then
                        System.Windows.Forms.Application.DoEvents()
                        Player_Renamed.Hjerter = 1
                        Lose(g)
                        LoadLevel(g, Player_Renamed.Level)

                        Debug.Print("Player killed")
                    End If
                End If
            End If
        End If

ABC:
        If Not Objekter_Renamed(A).Tag = 5 And Not Objekter_Renamed(A).Tag = 6 And Not Objekter_Renamed(A).Tag = 7 Then
            For B = 0 To 100
                If Objekter_Renamed(B).Enabled = True And Not B = A And Not Objekter_Renamed(B).Tag = Objekter_Renamed(A).Tag Then
                    If CCOLLIDE(Objekter_Renamed(B).X, Objekter_Renamed(B).Y, Objekter_Renamed(A).X, Objekter_Renamed(A).Y, BW) = 1 Then
                        If Objekter_Renamed(A).Vei = 0 Then Objekter_Renamed(A).Vei = 1 : GoTo AA
                        If Objekter_Renamed(A).Vei = 1 Then Objekter_Renamed(A).Vei = 0 : GoTo AA
                    End If
AA:
                End If
            Next
        End If

        If Objekter_Renamed(A).Tag = 1 Or Objekter_Renamed(A).Tag = 4 Then
            XX = Objekter_Renamed(A).X / BW
            YY = Objekter_Renamed(A).Y / BH
            If XX <= 0 Or YY < 0 Or XX > E(0) + 1 Or YY + 1 > E(1) Then Objekter_Renamed(A).Enabled = False : Exit Function

            If Colisjon(g, Objekter_Renamed(A).X, Objekter_Renamed(A).Y + 4 * elapsedTicks, False, "Down") = 0 Then
                Objekter_Renamed(A).Y = Objekter_Renamed(A).Y + 4 * elapsedTicks
            End If

            If COLLIDE(Player_Renamed.X, Player_Renamed.Y, Objekter_Renamed(A).X, Objekter_Renamed(A).Y, BW, "A") = 1 Then
                If Objekter_Renamed(A).Vei = 0 Then
                    If Player_Renamed.X > Objekter_Renamed(A).X Then Objekter_Renamed(A).Vei = 1
                    If Player_Renamed.X < Objekter_Renamed(A).X Then Objekter_Renamed(A).Vei = 2
                Else
                    If Player_Renamed.Uddelig <= 0 Then Player_Renamed.Hjerter = Player_Renamed.Hjerter - 1
                    Player_Renamed.Uddelig = 50

                    If Player_Renamed.Hjerter = 0 Then
                        System.Windows.Forms.Application.DoEvents()
                        Player_Renamed.Hjerter = 1
                        Lose(g)
                        LoadLevel(g, Player_Renamed.Level)

                        Debug.Print("Player killed")
                    End If
                End If
            End If

            If Objekter_Renamed(A).Vei = 1 Then
                If Colisjon(g, Objekter_Renamed(A).X - (PixelGo * 2 * elapsedTicks), Objekter_Renamed(A).Y, False, "Left") = 0 Then
                    Objekter_Renamed(A).X = Objekter_Renamed(A).X - (PixelGo * 2 * elapsedTicks)
                Else
                    Objekter_Renamed(A).Vei = 2
                End If
            End If

            If Objekter_Renamed(A).Vei = 2 Then
                If Colisjon(g, Objekter_Renamed(A).X + (PixelGo * 2 * elapsedTicks), Objekter_Renamed(A).Y, False, "Right") = 0 Then
                    Objekter_Renamed(A).X = Objekter_Renamed(A).X + (PixelGo * 2 * elapsedTicks)
                Else
                    Objekter_Renamed(A).Vei = 1
                End If
            End If
        End If

        ' Hoppende lava ting:
        If Objekter_Renamed(A).Tag = 7 Then
            Objekter_Renamed(A).A = Objekter_Renamed(A).A + elapsedTicks
            If COLLIDE(Player_Renamed.X, Player_Renamed.Y, Objekter_Renamed(A).X, Objekter_Renamed(A).Y, BW, "A") = 1 Then
                Lose(g)
                LoadLevel(g, Player_Renamed.Level)
                DoObjekts = 2

                Debug.Print("Player killed")
                Exit Function
            End If
            If Objekter_Renamed(A).Vei = 0 And Objekter_Renamed(A).A >= 40 Then Objekter_Renamed(A).A = 0 : Objekter_Renamed(A).Vei = 1
            If Objekter_Renamed(A).Vei = 1 Then Objekter_Renamed(A).Y = Objekter_Renamed(A).Y - PixelGo * elapsedTicks
            If Objekter_Renamed(A).Vei = 2 Then Objekter_Renamed(A).Y = Objekter_Renamed(A).Y + PixelGo * elapsedTicks
            If Objekter_Renamed(A).Vei = 1 Then
                If Objekter_Renamed(A).A >= 20 Or Colisjon(g, Objekter_Renamed(A).X, Objekter_Renamed(A).Y, False, "Up") = 1 Then
                    Objekter_Renamed(A).Vei = 2
                    Objekter_Renamed(A).A = 20
                End If
            End If

            If Objekter_Renamed(A).Vei = 2 And Objekter_Renamed(A).A >= 40 Then
                Objekter_Renamed(A).Y = Objekter_Renamed(A).Y + PixelGo * elapsedTicks
                Objekter_Renamed(A).Vei = 0
                Objekter_Renamed(A).A = 0
            End If
        End If


    End Function

    Sub LoadLevel(ByVal g As Graphics, ByRef Level As Short)

        Dim A As Integer
        Dim B, D As Short
        Dim C As String

        A = FreeFile()

        Debug.Print("Open level " & Level)

        XXX = 0
        YYY = 0
        Player_Renamed.Ok = True
        DeleteAllObjekts()

        FileOpen(A, GetLevelPath("Level" + CStr(Level) & ".dat"), OpenMode.Input)

        Input(A, E(0))
        Input(A, E(1))

        ReDim Bokser(E(0), E(1))

        Do Until EOF(A)
            C = LineInput(A)
            For D = 1 To Len(C)
                Bokser(D - 1, B) = Mid(C, D, 1)
                If Bokser(D - 1, B) = "Y" Then
                    Bokser(D - 1, B) = "A"
                    Player_Renamed.X = (D - 1) * BW
                    Player_Renamed.Y = B * BH
                End If
                If Bokser(D - 1, B) = "N" Then CreateObjekt((D - 1) * BW, B * BH, 0) : Bokser(D - 1, B) = "A"
                If Bokser(D - 1, B) = "O" Then CreateObjekt((D - 1) * BW, B * BH, 2) : Bokser(D - 1, B) = "A"
                If Bokser(D - 1, B) = "P" Then CreateObjekt((D - 1) * BW, B * BH, 3) : Bokser(D - 1, B) = "A"
                If Bokser(D - 1, B) = "3" Then CreateObjekt((D - 1) * BW, B * BH, 7) : Bokser(D - 1, B) = "1"
            Next
            B = B + 1
        Loop

        FileClose(A)

        If Not Player_Renamed.XX = 0 And Not Player_Renamed.YY = 0 Then
            Player_Renamed.X = Player_Renamed.XX * BW
            Player_Renamed.Y = Player_Renamed.YY * BH
        End If

        Debug.Print("Teleport player to X: " & Player_Renamed.X & ", Y: " & Player_Renamed.Y)

        If Not g Is Nothing Then
            DrawAll(g)
        End If

        LevelJustLoaded = True

    End Sub

    Sub DrawAll(ByVal g As Graphics)
        DrawAll(g, Nothing, GetGameState(), 0)
    End Sub

    Sub DrawAll(ByVal g As Graphics, prevState As GameState, currState As GameState, deltaTicks As Double)

        Dim B, A, D As Short

        g.FillRectangle(New SolidBrush(BackColor), ScreenRect)
        'Buff.BltColorFill(ViewRect, BackColor)

        Dim gridOffsetX As Integer = LinearInterpolation(prevState, currState, deltaTicks, Function(s) s.offsetX)
        Dim gridOffsetY As Integer = LinearInterpolation(prevState, currState, deltaTicks, Function(s) s.offsetY)

        For B = 0 To E(1)
            For A = Int((0 - gridOffsetX) / BW) To Int((0 - gridOffsetX) / BW) + 20
                DrawBoks(g, A, B, gridOffsetX, gridOffsetY)
            Next
        Next

        Dim playerX As Integer = LinearInterpolation(prevState, currState, deltaTicks, Function(s) s.player.X)
        Dim playerY As Integer = LinearInterpolation(prevState, currState, deltaTicks, Function(s) s.player.Y)

        If currState.player.Uddelig <= 0 Then
            If currState.player.Vei = False Then BitmapBlt(g, gridOffsetX + playerX, gridOffsetY + playerY, 18, 32, 0, 0, 32, vbSrcAnd)
            If currState.player.Vei = True Then BitmapBlt(g, gridOffsetX + playerX, gridOffsetY + playerY, 18, 32, 0, 56, 32, vbSrcAnd)
        Else
            If currState.player.Vei = False Then BitmapBlt(g, gridOffsetX + playerX, gridOffsetY + playerY, 18, 32, 0, 0, 96, vbSrcAnd)
            If currState.player.Vei = True Then BitmapBlt(g, gridOffsetX + playerX, gridOffsetY + playerY, 18, 32, 0, 9, 96, vbSrcAnd)
        End If

        BitmapBlt(g, ScreenWidth - TextWidth(g, GameFont, "        100"), 2, 18, 32, 0, 48, 32, vbSrcAnd)
        BitmapBlt(g, 0, 2, 18, 32, 0, 0, 32, vbSrcAnd)

        g.DrawString(CStr(Penger), GameFont, Brushes.Black, ScreenWidth - TextWidth(g, GameFont, "     100"), 0)
        g.DrawString(" x " & currState.player.Liv, GameFont, Brushes.Black, 0, 12)

        For A = 1 To 6
            If currState.player.Hjerter >= A Then
                If currState.player.MaxH >= A Then BitmapBlt(g, 100 + (40 + ((A - 1) * 32)), 2, 32, 32, 0, 64, 80, vbSrcAnd)
            Else
                If currState.player.MaxH >= A Then BitmapBlt(g, 100 + (40 + ((A - 1) * 32)), 2, 32, 32, 0, 48, 80, vbSrcAnd)
            End If
        Next

        For A = 0 To 999
            ' Skip objects that are not enabled
            If Not currState.objects(A).Enabled Then
                Continue For
            End If

            Dim hasPrevState = prevState IsNot Nothing AndAlso prevState.objects(A).Enabled

            Dim objectTag As Short = currState.objects(A).Tag
            Dim objectOrientation As Short = currState.objects(A).Vei
            Dim objectX As Double = LinearInterpolation(If(hasPrevState, prevState, Nothing), currState, deltaTicks, Function(s) s.objects(A).X)
            Dim objectY As Double = LinearInterpolation(If(hasPrevState, prevState, Nothing), currState, deltaTicks, Function(s) s.objects(A).Y)

            If objectX + 36 > 0 - gridOffsetX And objectX - 36 < (0 - gridOffsetX) + 640 Then
                Select Case objectTag
                    Case 0
                        If objectOrientation = 1 Then BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 16, 32, vbSrcAnd)
                        If objectOrientation = 0 Then BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 32, 32, vbSrcAnd)
                    Case 1
                        BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 30, 32, 0, 65, 32, vbSrcAnd)
                    Case 2
                        BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 16, 48, vbSrcAnd)
                    Case 3
                        If objectOrientation = 1 Then BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 48, 48, vbSrcAnd)
                        If objectOrientation = 0 Then BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 0, 64, vbSrcAnd)
                    Case 4
                        BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 32, 64, vbSrcAnd)
                    Case 5
                        BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 16, 80, vbSrcAnd)
                    Case 6
                        BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 32, 112, vbSrcAnd)
                    Case 7
                        If objectOrientation = 1 Then BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 32, 128, vbSrcAnd)
                        If objectOrientation = 2 Then BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 64, 128, vbSrcAnd)
                    Case 8
                        BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 16, 144, vbSrcAnd)
                    Case 9
                        BitmapBlt(g, gridOffsetX + objectX, gridOffsetY + objectY, 32, 32, 0, 32, 144, vbSrcAnd)
                End Select
            End If
        Next

        g.DrawString(CStr(H), GameFont, Brushes.Black, 0, TextHeight(g, GameFont, "A"))

    End Sub

    Sub DrawBoks(ByVal g As Graphics, ByVal X As Short, ByVal Y As Short, ByVal XXX As Integer, ByVal YYY As Integer)

        If X < 0 Or Y < 0 Or X > E(0) Or Y > E(1) Then Exit Sub

        If Bokser(X, Y) = "B" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 0, 0, vbSrcCopy)
        If Bokser(X, Y) = "C" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 0, vbSrcCopy)
        If Bokser(X, Y) = "D" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 32, 0, vbSrcCopy)
        If Bokser(X, Y) = "E" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 48, 0, vbSrcCopy)
        If Bokser(X, Y) = "F" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 32, 16, vbSrcCopy)
        If Bokser(X, Y) = "G" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 48, 16, vbSrcCopy)
        If Bokser(X, Y) = "H" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 0, 16, vbSrcCopy)
        If Bokser(X, Y) = "I" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, CInt(State) * 16, 0, vbSrcAnd, 1)
        If Bokser(X, Y) = "J" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 16, vbSrcCopy)
        If Bokser(X, Y) = "K" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 16, vbSrcCopy)
        If Bokser(X, Y) = "L" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 16, vbSrcCopy)
        If Bokser(X, Y) = "M" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 64, 16, vbSrcCopy)
        If Bokser(X, Y) = "Q" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 0, 80, vbSrcCopy)
        If Bokser(X, Y) = "R" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 0, 80, vbSrcCopy)
        If Bokser(X, Y) = "S" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 0, 80, vbSrcCopy)
        If Bokser(X, Y) = "T" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 32, 96, vbSrcCopy)
        If Bokser(X, Y) = "U" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 48, 96, vbSrcCopy)
        If Bokser(X, Y) = "V" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 64, 96, vbSrcCopy)
        If Bokser(X, Y) = "W" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 0, 112, vbSrcCopy)
        If Bokser(X, Y) = "X" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 112, vbSrcCopy)
        If Bokser(X, Y) = "0" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 64, 112, vbSrcCopy)
        If Bokser(X, Y) = "1" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 0, 128, vbSrcCopy)
        If Bokser(X, Y) = "2" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 128, vbSrcCopy)
        If Bokser(X, Y) = "4" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 16, vbSrcCopy)
        If Bokser(X, Y) = "5" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 16, 16, vbSrcCopy)
        If Bokser(X, Y) = "6" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 48, 112, vbSrcCopy)
        If Bokser(X, Y) = "7" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 48, 128, vbSrcCopy)
        If Bokser(X, Y) = "Z" Then BitmapBlt(g, XXX + (X * BW), YYY + (Y * BH), BW, BH, 0, 64, 64, vbSrcAnd)

    End Sub


    Function COLLIDE(ByVal XPOSITION1 As Short, ByVal YPOSITION1 As Short, ByVal XPOSITION2 As Short, ByVal YPOSITION2 As Short, ByVal BOXSIZE As Short, ByVal Ting As String) As Short

        ABFG = BOXSIZE
        If Ting = "1" Then ABFG = 8

        If XPOSITION1 > XPOSITION2 - 28 And XPOSITION1 < XPOSITION2 + 16 And YPOSITION1 > YPOSITION2 - BOXSIZE And YPOSITION1 < YPOSITION2 + ABFG Then COLLIDE = 1

    End Function

    Function CCOLLIDE(ByVal XPOSITION1 As Short, ByVal YPOSITION1 As Short, ByVal XPOSITION2 As Short, ByVal YPOSITION2 As Short, ByVal BOXSIZE As Short) As Short

        If XPOSITION1 > XPOSITION2 - BOXSIZE And XPOSITION1 < XPOSITION2 + BOXSIZE And YPOSITION1 > YPOSITION2 - BOXSIZE And YPOSITION1 < YPOSITION2 + BOXSIZE Then CCOLLIDE = 1

    End Function

    Function Colisjon(ByVal g As Graphics, ByVal XX As Short, ByVal YY As Short, ByVal Play As Boolean, ByVal Command_Renamed As String) As Short

        Dim BufferY, BufferX, D, B, A, C, F, BufferXX, BufferYY As Short

        Select Case LCase(Command_Renamed)
            Case "left" : BufferX = Int(XX / BW) - 1 : BufferXX = Int(XX / BW) : BufferY = Int(YY / BH) - 1 : BufferYY = Int(YY / BH) + 1
            Case "right" : BufferX = Int(XX / BW) : BufferXX = Int(XX / BW) + 1 : BufferY = Int(YY / BH) - 1 : BufferYY = Int(YY / BH) + 1
            Case "up" : BufferX = Int(XX / BW) - 1 : BufferXX = Int(XX / BW) + 1 : BufferY = Int(YY / BH) - 1 : BufferYY = Int(YY / BH)
            Case "down" : BufferX = Int(XX / BW) - 1 : BufferXX = Int(XX / BW) + 1 : BufferY = Int(YY / BH) : BufferYY = Int(YY / BH) + 1
            Case Else : BufferX = Int(XX / BW) - 1 : BufferXX = Int(XX / BW) + 1 : BufferY = Int(YY / BH) - 1 : BufferYY = Int(YY / BH) + 1
        End Select

        For A = BufferX To BufferXX
            For B = BufferY To BufferYY
                If A < 0 Or B < 0 Or A > E(0) + 1 Or B > E(1) Then GoTo A
                If Not Bokser(A, B) = "A" And Not Bokser(A, B) = "T" And Not Bokser(A, B) = "U" And Not Bokser(A, B) = "W" Then
                    If Play = False And Bokser(A, B) = "I" Then GoTo A
                    If COLLIDE(A * BW, B * BH, XX, YY, BW, Bokser(A, B)) = 1 Then
                        C = DoHO(g, A, B, Play)
                        If C = 1 Then GoTo A
                        If C = 2 Then Colisjon = 2 : Exit Function
                        Colisjon = 1
                        Exit Function
                    End If
                End If
A:
            Next
        Next

    End Function

    Function DoHO(ByVal g As Graphics, ByRef A As Short, ByRef B As Short, ByRef Play As Boolean) As Short

        If Bokser(A, B) = "I" And Play = True Then DoHO = 1 : PlayWAV("jingle.wav") : Bokser(A, B) = "A" : Penger = Penger + 1
        If Bokser(A, B) = "Z" And Play = True Then
            DoHO = 1
            Player_Renamed.Level = Player_Renamed.Level + 1
            Player_Renamed.XX = 0
            Player_Renamed.YY = 0
            Player_Renamed.X = 0
            Player_Renamed.Y = 0
            LoadLevel(g, Player_Renamed.Level)
            If GameOver Then Exit Function
            DoHO = 2
            Exit Function
        End If

        If Bokser(A, B) = "V" And Play = True Then
            If Play = True Then DoHO = 1 : Player_Renamed.XX = A : Player_Renamed.YY = B : Bokser(A, B) = "W" Else DoHO = 1 : Exit Function
        End If

        If Bokser(A, B) = "0" Or Bokser(A, B) = "1" Or Bokser(A, B) = "2" Then
            If Play = True Then
                Lose(g)
                DoHO = 2
                LoadLevel(g, Player_Renamed.Level)

                Debug.Print("Player killed")
                Exit Function
            End If
            If Play = False Then DoHO = 1
        End If

    End Function

    Sub Dunk()

        Dim XX, YY As Short

        XX = Player_Renamed.X / BW
        YY = Player_Renamed.Y / BH

        If Bokser(XX, YY - 1) = "K" Or Bokser(XX, YY - 1) = "S" Then PlayWAV("box.wav") : Bokser(XX, YY - 1) = "M" : CreateObjekt(XX * BW, (YY - 2) * BH, 5) : GoTo A
        If Bokser(XX, YY - 1) = "J" Or Bokser(XX, YY - 1) = "R" Then PlayWAV("box.wav") : Bokser(XX, YY - 1) = "M" : Penger = Penger + 1 Else If Not Bokser(XX, YY - 1) = "Q" Then PlayWAV("Nothing.wav")
        If Bokser(XX, YY - 1) = "Q" Then Bokser(XX, YY - 1) = "A"
        If Bokser(XX, YY - 1) = "L" Then Bokser(XX, YY - 1) = "M" : CreateObjekt(XX * BW, (YY - 2) * BH, 6) : PlayWAV("box.wav") : GoTo A
        If Bokser(XX, YY - 1) = "4" Then Bokser(XX, YY - 1) = "M" : CreateObjekt(XX * BW, (YY - 2) * BH, 8) : PlayWAV("box.wav") : GoTo A
        If Bokser(XX, YY - 1) = "5" Then Bokser(XX, YY - 1) = "M" : CreateObjekt(XX * BW, (YY - 2) * BH, 9) : PlayWAV("box.wav") : GoTo A
A:

    End Sub

    Public Sub PlayWAV(ByRef WAVtoPlay As String)

        On Error Resume Next
        If Sound = False Then Exit Sub

        LoadWAV(WAVtoPlay)

        Dim fullPath As String = GetResourcePath(WAVtoPlay)
        mciSendString("play " & Chr(34) & fullPath & Chr(34) & " from 0", Nothing, 0, 0)

    End Sub

    Public Sub LoadWAV(ByVal WAVtoPlay As String)

        Dim fullPath As String = GetResourcePath(WAVtoPlay)

        If Not soundCache.ContainsKey(fullPath) Then
            soundCache(WAVtoPlay) = True
            mciSendString("open " & Chr(34) & fullPath & Chr(34), Nothing, 0, 0)
        End If

    End Sub

    Public Sub UnloadWAV()
        ' Clear sounds
        For Each soundFile As String In soundCache.Keys
            mciSendString("close " & Chr(34) & soundFile & Chr(34), Nothing, 0, 0)
        Next
    End Sub

    Public Sub DeleteAllObjekts()

        Dim A As Object
        For A = 0 To 999
            Objekter_Renamed(A).Enabled = False
        Next

    End Sub

    Public Sub CreateObjekt(ByRef X As Short, ByRef Y As Short, ByRef Tag As Short)

        Dim A, Index As Object
        For A = 0 To 999
            If Objekter_Renamed(A).Enabled = False Then
                Index = A : Exit For
            End If
        Next

        Objekter_Renamed(Index).Enabled = True
        Objekter_Renamed(Index).X = X
        Objekter_Renamed(Index).Y = Y
        Objekter_Renamed(Index).Vei = 0
        Objekter_Renamed(Index).Tag = Tag

    End Sub

    Sub Lose(ByVal g As Graphics)

        Dim A As Short
        Player_Renamed.Liv = Player_Renamed.Liv - 1
        A = -8

        Do Until Player_Renamed.Y > ScreenHeight
            DrawAll(g)

            If Not A = 8 Then A = A + 1
            Player_Renamed.Y = Player_Renamed.Y + A

            System.Windows.Forms.Application.DoEvents()
            If Not Speed = 0 Then Thread.Sleep(Speed)
        Loop

    End Sub

    Sub PlussPLiv()

        Player_Renamed.Hjerter = Player_Renamed.Hjerter + 1
        If Player_Renamed.Hjerter = Player_Renamed.MaxH + 1 Then Player_Renamed.Hjerter = Player_Renamed.MaxH

    End Sub

    Sub PlussPMax()
        If Not Player_Renamed.MaxH = 6 Then Player_Renamed.MaxH = Player_Renamed.MaxH + 1
    End Sub

    Sub BitmapBlt(ByRef g As Graphics, ByVal X As Short, ByVal Y As Short, ByVal nWidth As Short, ByVal nHeight As Short, ByVal hSrcDC As Integer, ByVal XSrc As Short, ByVal YSrc As Short, ByVal dwRop As Integer, Optional ByVal picIndex As Integer = 0)

        'If X - nWidth > ScreenRect.Width Then Exit Sub
        'If Y - nHeight > ScreenRect.Height Then Exit Sub

        If X < 0 Then
            If X < -nWidth Then Exit Sub
            XXXX = 0 - X
        Else
            XXXX = 0
        End If

        If Y < 0 Then
            If Y < -nHeight Then Exit Sub
            YYYY = 0 - Y
        Else
            YYYY = 0
        End If

        If X + nWidth > ScreenRect.Right Then XY = ScreenRect.Right - (X + nWidth) Else XY = 0

        Dim sourceRect As Rectangle = Rectangle.FromLTRB(
            (XSrc * 2) + XXXX,
            (YSrc * 2) + YYYY,
            (XSrc * 2) + nWidth + XY,
            (YSrc * 2) + nHeight)

        Dim destRect As New Rectangle(X + XXXX, Y + YYYY, sourceRect.Width, sourceRect.Height)

        g.DrawImage(Pic(picIndex), destRect, sourceRect.X, sourceRect.Y, sourceRect.Width, sourceRect.Height, GraphicsUnit.Pixel)

    End Sub

    Sub TaFraLiv(ByVal g As Graphics)

        Player_Renamed.Hjerter = Player_Renamed.Hjerter - 1
        If Player_Renamed.Hjerter = 0 Then Player_Renamed.Hjerter = 1 : LoadLevel(g, Player_Renamed.Level)

    End Sub

    Function LinearInterpolation(prevState As GameState, currentState As GameState, deltaTicks As Double, stateExtractor As IntegerExtractor) As Integer
        ' If no previous state exists, just return current state
        If prevState Is Nothing Then
            Return stateExtractor(currentState)
        End If

        ' Otherwise, interpolate between the previous state and current state
        Dim t As Double = Math.Min(Math.Max(deltaTicks, 0), 1) ' Restrict to 0 - 1 
        Return stateExtractor(prevState) * (1 - t) + stateExtractor(currentState) * t

    End Function

    Function LinearInterpolation(prevState As GameState, currentState As GameState, deltaTicks As Double, stateExtractor As DoubleExtractor) As Double
        ' If no previous state exists, just return current state
        If prevState Is Nothing Then
            Return stateExtractor(currentState)
        End If

        ' Otherwise, interpolate between the previous state and current state
        Dim t As Double = Math.Min(Math.Max(deltaTicks, 0), 1) ' Restrict to 0 - 1 
        Return stateExtractor(prevState) * (1 - t) + stateExtractor(currentState) * t

    End Function
End Module